//  Wheel.h - a wheel.
//
//  Copyright (C) 2001--2002 Sam Varner
//
//  This file is part of Vamos Automotive Simulator.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef _WHEEL_H_
#define _WHEEL_H_

#include <vamos/body/Suspension.h>
#include <vamos/body/Tire.h>
#include <vamos/body/Brake.h>
#include <vamos/body/Contact_Point.h>
#include <vamos/geometry/Surface.h>
#include <vamos/geometry/Three_Vector.h>
#include <vamos/geometry/Inertia_Tensor.h>

using Vamos_Geometry::Three_Vector;

namespace Vamos_Body
{
  //* A wheel.  
  // Handles wheel geometry.  See Tire, Brake and Suspension for
  // related functionality.
  class Wheel : public Contact_Point
  {
  private:

	// The initial position of the wheel.
	Three_Vector m_original_position;

	// The suspension that the wheel is attached to.
	Suspension* mp_suspension;

	// The tire that is attached to the wheel.
	Tire m_tire;

	// The brake for the wheel.
	Brake m_brake;

	// The velocity of the ground relative to the wheel.
	Three_Vector m_ground_velocity;

	// The current normal vector for the road.
	Three_Vector m_normal;

	// The angular velocity of the wheel.
 	Three_Vector m_ang_velocity;
	
	// The surface we're currently on
	Vamos_Geometry::Surface m_surface;

	// The torque applied to the tire.
	double m_tire_torque;

	double m_braking_torque;

	Three_Vector m_wheel_force;
	
  public:
	//** Constructor
	Wheel (double mass, 
		   Three_Vector position, 
		   double restitution,
		   Suspension* suspension, 
		   const Tire& tire, 
		   const Brake& brake);

	// Find and store the forces and torques for the current
	// configuration.
	void find_forces ();

	// Advance the wheel forward in time by TIME.
	void propagate (double time);

	// Undo the last propagation.
	void rewind ();

 	// Handle hard collisions.  The return value is the displacement
	// of the particle as a result of the contact.  This function is
	// called by a Body object.
	Three_Vector contact (const Three_Vector& position,
						  const Vamos_Geometry::Inertia_Tensor& inertia,
						  const Three_Vector& vecocity, 
						  const Three_Vector& distance,
						  const Three_Vector& ang_velocity,
						  const Vamos_Geometry::Surface& surface);

	// Handle soft collisions.  The return value is the displacement
	// of the particle as a result of the contact.  This function is
	// called by a Car object.
	Three_Vector wheel_contact (const Three_Vector& position,
								const Vamos_Geometry::Inertia_Tensor& inertia,
								const Three_Vector& velocity, 
								const Three_Vector& distance,
								const Three_Vector& ang_velocity,
								const Vamos_Geometry::Surface& surface);

	// Apply the torque TORQUE_IN to the tire.  This torque results
	// from acceleration or braking.
	void torque (double torque_in);

	// Return the torque exerted on the wheel by the brake when a
	// pressure of FACTOR * Brake::m_max_pressure is applied to the
	// brake.
	double brake_torque (double factor);

	// Set the steering angle.
	void steer (double degree_angle) { mp_suspension->steer (degree_angle); }


	// Return the speed of the contact patch of the tire with respect
	// to the ground.
	double slide () const { return m_tire.slide (); }

	// Return the wheel's rotational speed in radians per second.
	double rotational_speed () const { return m_tire.rotational_speed (); }

	// Return the linear speed of the tire tread.  This is the speed
	// that a speedometer on this wheel would read.
	double speed () const { return m_tire.speed (); }

	// Return the position relative to the body where the wheel exerts
	// its force.
	Three_Vector force_position () const;

	// Return the position relative to the body where the wheel exerts
	// its torque.
	Three_Vector torque_position () const;

	// Return the position of the wheel relative to the body for the
	// purpose of detecting collisions.
	Three_Vector contact_position () const;

	// Return the slip ratio for the wheel;
	double slip () const;

	// Return the wheel to its initial conditions.
	void reset ();
  };
}

#endif // !_WHEEL_H_
