//	Vamos Automotive Simulator
//  Copyright (C) 2001--2002 Sam Varner
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef _ROAD_SEGMENT_H_
#define _ROAD_SEGMENT_H_

#include <vamos/geometry/Spline.h>
#include <vamos/geometry/Three_Vector.h>
#include <vamos/geometry/Two_Point.h>
#include <vamos/geometry/Conversions.h>
#include <vamos/geometry/Surface.h>

#include <GL/glu.h>

#include <string>
#include <vector>
#include <cmath>

using Vamos_Geometry::Three_Vector;

namespace Vamos_Track
{
  class Not_On_This_Segment
  {
	// How far off the far end (positive) or near end (negative) of the 
	// current segment the point is during a call to track_coordinates.
	double m_difference;
  public:
	Not_On_This_Segment (double diff) : m_difference (diff) {};
	double difference () const { return m_difference; }
  };

  class Road_Segment
  {
  protected:
	GLuint* m_gl_texture_name;
	GLuint m_gl_list_id;

	// The total length of the segment.
	double m_length;
	// The length of a subdivision of the segment.
	double m_resolution;

	// The endpoints of the segment.
	Three_Vector m_start_coords;
	Three_Vector m_end_coords;
	double m_left;
	double m_right;
	double m_start_angle;
	double m_start_distance;

	Three_Vector m_current_coords;
	double m_current_distance;
	double m_current_angle;
	bool m_on_left_side;

	// True when we're rendering the first subdivision.
	bool m_first;
	// True when we're rendering the last subdivision.
	bool m_last;

	std::vector <Vamos_Geometry::Two_Point> m_elevation_points;
	double m_start_elevation;
	double m_end_elevation;
	double m_end_pitch;
	double m_elevation_inc;
	Vamos_Geometry::Spline* mp_elevation_curve;

	std::vector <Vamos_Geometry::Two_Point> m_bank_points;
	double m_start_bank;
	double m_end_bank;
	double m_end_twist;
	double m_bank_pivot;
	double m_bank_inc;
	Vamos_Geometry::Spline* mp_bank_curve;

	std::vector <Vamos_Geometry::Surface*> m_surfaces;

	double m_inside_width;
	double m_outside_width;

	double bank_height (double along, double from_center) const;

	// True if this is the last segment of a circuit.
	bool m_last_segment;

	// bounding dimensions
	double m_min_x;
	double m_max_x;
	double m_min_y;
	double m_max_y;

	virtual void start (const Three_Vector& start_corrds, 
						double left, double right, double angle,
						double start_elevation, double start_pitch,
						double start_bank, double start_twist) ;

  public:
	Road_Segment (double resolution, double length, 
				  const std::vector <Vamos_Geometry::Two_Point>& 
				  elevation_points,
				  double end_elevation, double end_pitch,
				  const std::vector <Vamos_Geometry::Two_Point>& bank_points,
				  double end_bank, double end_twist, double bank_pivot,
				  const std::vector <Vamos_Geometry::Surface*>& surfaces);
	Road_Segment (const Road_Segment& seg);
	virtual ~Road_Segment ();

	double length () const { return m_length; }
	// Change the length of a segment.  The track must be re-built when this
	// is done.
	void length (double new_length) { m_length = new_length; }
	double resolution () const { return m_resolution; }
	virtual double end_angle () const { return m_start_angle; }
	virtual Three_Vector coordinates (const Three_Vector& world_pos) const = 0;

	virtual Three_Vector next_vertex () = 0;
	bool finished () const { return m_last; }
	Three_Vector end_coords () const { return m_end_coords; } 
	Three_Vector start_coords () const { return m_start_coords; } 

	double end_elevation () const { return m_end_elevation; }
	double end_pitch () const { return m_end_pitch; }

	double end_bank () const { return m_end_bank; }
	double end_twist () const { return m_end_twist; }

	double elevation (double, double) const;
	virtual double angle (double along) const = 0;
	virtual Three_Vector normal (double along, double from_center) const = 0;
	virtual Three_Vector barrier_normal (double along, 
										 double from_center) const = 0;

	void build (double start_distance,
				const Three_Vector& start_coords, double start_angle,
				double start_elevation, double start_pitch,
				double start_bank, double start_twist);
	void draw () const;

	const std::vector <Vamos_Geometry::Surface*>& surfaces () const 
	{ return m_surfaces; }

	double inside_width () const { return m_inside_width; }
	double outside_width () const { return m_outside_width; }
	void last_segment (bool last) { m_last_segment = last; }
	double start_distance () const { return m_start_distance; }

	double min_x () const { return m_min_x; }
	double max_x () const { return m_max_x; }
	double min_y () const { return m_min_y; }
	double max_y () const { return m_max_y; }
  };

  //===========================================================================
  class Straight_Road : public Road_Segment
  {
  public:
	Straight_Road (double resolution, double length, 
				   const std::vector <Vamos_Geometry::Two_Point>& 
				   elevation_points,
				   double end_elevation, double end_pitch,
				   const std::vector <Vamos_Geometry::Two_Point>& bank_points,
				   double end_bank, double end_twist, double bank_pivot,
				   const std::vector <Vamos_Geometry::Surface*>& surfaces);

	Three_Vector coordinates (const Three_Vector& world_pos) const;
	Three_Vector next_vertex ();
	double angle (double along) const { return m_start_angle; }
	Three_Vector normal (double, double) const;
	Three_Vector barrier_normal (double along, double from_center) const;
  };

  //===========================================================================
  class Arc_Road : public Road_Segment
  {
	double m_radius;
	double m_arc;

	double m_center_x;
	double m_center_y;

	void start (const Three_Vector& start_coords, 
				double left, double right, double angle,
				double start_elevation, double start_pitch,
				double start_bank, double start_twist);

  public:
	enum Arc_Type
	{
	  LENGTH_AND_RADIUS,
	  LENGTH_AND_ANGLE,
	  RADIUS_AND_ANGLE
	};

	Arc_Road (double resolution, double arg1, double arg2,
			  const std::vector <Vamos_Geometry::Two_Point>& elevation_points,
			  double end_elevation, double end_pitch,
			  const std::vector <Vamos_Geometry::Two_Point>& bank_points,
			  double end_bank, double end_twist, double bank_pivot,
			  const std::vector <Vamos_Geometry::Surface*>& surfaces,
			  Arc_Type type = LENGTH_AND_RADIUS);

	Three_Vector coordinates (const Three_Vector& world_pos) const;

	Three_Vector next_vertex ();
	double end_angle () const { return m_start_angle + m_arc; }
	double angle (double along) const;
	Three_Vector barrier_normal (double along, double from_center) const;
	double radius () const { return m_radius; }
	Three_Vector normal (double, double) const;
  };
}

#endif
